; ******************************************************
; This example ilustrates the use of SPM instruction
; to self-program the AVR.
; It fills a flash page with NOP instructions, where
; originally, some RETI opcodes were written.
;
; Bootload code adapted from Atmel's example Mega8 manual,
; page 213
; ******************************************************

.include "..\include\m8def.inc"

; Register based variables
;
.def temp1 =r16
.def temp2 =r17
.def spmcrval =r18
.def looplo =r24
.def loophi =r25

; Flash page related definitions. The example will be
; performed in Flash page #16
; PAGESIZE definition (from .inc file) is in WORDS
;
.equ PAGESIZEB = PAGESIZE * 2  ; Size in BYTES
.equ PAGE_16 = PAGESIZE * 16   ; The page to be changed
.equ PAGE_16B = PAGE_16 * 2    ; In byte addrress

; *****************************************************
; Normal reset vector, not used, since the BOOTRST fuse
; has been set to 0 in the Project File. Change the
; fuse value, or comment the fuses definition string in the
; Project File to execute this code.
;
.ORG 0x0000
Normal_reset:
	rjmp dummy
	nop          ; This piece of code
	nop          ; never gets executed
	nop          ; unless you change the
	nop          ; BOOTRST fuse value
dummy:
	rjmp dummy	

; *****************************************************
;  This code page will be self-modified by the bootloader
;
.ORG PAGE_16
Code_changed:
	reti         ; These pack of RETI instructions will
	reti         ; be changed by NOP codes (0x0000)
	reti         ;
	reti
	reti         ; A full page of NOPs will be
	reti         ; inserted (32 words)
	reti
	reti         ; See how the Stop/X-ref buttons
	reti         ; are removed, sine after bootloading
	reti         ; these instructions are no longer
	reti         ; RETI but NOP, so the debug information is
   reti         ; no longer valid. Follow the code in Program
   				 ; Memory window.

.ORG PAGE_16 + PAGESIZE
forever:        ; Start of the next page. Not touched by bootloader
	rjmp forever ;	Program flow will end here after bootloading

; ******************************************************
; Fuses BOOTSZ and BOOTRST force the reset to be in the
; bootload area, in this case 0xF80.
;
.org SMALLBOOTSTART
Boot_reset:
 	ldi temp1, 0xDF   ; Init stack pointer
   out SPL, temp1

	ldi ZL, low(PAGE_16B)    ; Load in Z the address of page.
	ldi ZH, high(PAGE_16B)	 ; Must be BYTE address, so must be
						          ; multiplied by 2
						          						
   ; First write the flash page that will update the code
   ; and then jump to the new code, just loaded
   ;
	call Write_page
	rjmp Code_changed
	
; *****************************************************
;	Perform the necessary operations to perform a page write
;					          		
Write_page:

	; Page erase
	; ----------
	ldi spmcrval, (1<<PGERS) | (1<<SPMEN)
	call Do_spm
	
	; Re-enable the RWW section
	; -------------------------
	ldi spmcrval, (1<<RWWSRE) | (1<<SPMEN)
	call Do_spm
	
	; Fill Flash page buffer with NOP opcodes
	; ---------------------------------------
	ldi looplo, low(PAGESIZEB)  ; init loop variable
	ldi loophi, high(PAGESIZEB) ; not required for PAGESIZEB<=256
	
	ldi temp1, 0   ; NOP code = 0x0000, to be put in r1:r0
	mov r1, temp1  ; r1:r0 is used as source of data (16bits)
	mov r0, temp1  ; to load the Flash page buffe
	
Wrloop:
	ldi spmcrval, (1<<SPMEN)
	call Do_spm
	adiw ZH:ZL, 2
	sbiw loophi:looplo, 2    ; use subi for PAGESIZEB<=256
	brne Wrloop
	
	; Execute page write; page addres (in bytes) --> Z
	; --------------------------------------------------
	ldi ZL, low(PAGE_16B)
	ldi ZH, high(PAGE_16B)
	ldi spmcrval, (1<<PGWRT) | (1<<SPMEN)
	call Do_spm
	
	; Re-enable the RWW section
	; -------------------------
	ldi spmcrval, (1<<RWWSRE) | (1<<SPMEN)
	call Do_spm
	
	; Read back and check, optional
	; -----------------------------
	ldi looplo, low(PAGESIZEB)   ; Init loop variable
	ldi loophi, high(PAGESIZEB)  ; Not required for PAGESIZEB <=256
Rdloop:
	lpm r0, Z+
	ldi temp1, 0   ; NULL opcode, must be all zeroes
	mov r0, temp1
	cpse r0, r1
	jmp Error
	sbiw loophi:looplo, 1 			; Use subi for PAGESIZEB<=256
	brne Rdloop
	
	; Return to RWW section. Verify that RWW section is safe to read
	; --------------------------------------------------------------
Return:
	in temp1, SPMCR
	sbrs temp1, RWWSB ; If RWWSB is set, the RWW section is not ready yet
	ret
	
	; Re-enable the RWW section
	; -------------------------
	ldi spmcrval, (1<<RWWSRE) | (1<<SPMEN)
	call Do_spm
	rjmp Return

; *****************************************
; This routine performs the SPM instruction
;
Do_spm:
	; check for previous SPM complete
	; -------------------------------
Wait_spm:
	in temp1, SPMCR
	sbrc temp1, SPMEN
	rjmp Wait_spm
	
	; Input: spmcrval determines SPM action disable interrupts if enabled,
	; store status
	;
	in temp2, SREG
	cli
	
	; Check that no EEPROM write access is present
	; --------------------------------------------
	Wait_ee:
	sbic EECR, EEWE
	rjmp Wait_ee
	
	; SPM timed sequence
	; ------------------
	out SPMCR, spmcrval
	spm
	
	; Restore SREG (to enable interrupts if originally enabled)
   ; ---------------------------------------------------------
	out SREG, temp2
	ret

; *********************************
; Erorr handling, not implemented
;
Error:
	rjmp Error
	nop
	ret












