;******************************************************
; TWI example
;
; This application demonstrates the use of the TWI cell
; in combination with two I2C monitor cells placed in the
; Project File.
;
; How to put the Mega8 in its different TWI modes:
;
; * Slave receiver. The Mega8 micro is set by default as Slave
;   Receiver with address = 0x0E. To transmit data from a I2C
;   monitor, being picked up by the Mega8 slave, set 0E in the
;   'addr' box and the data in its corresponding box.(hex format,
;   separated by spaces, like OE 12 33 34 ....then, push "Master TX".

; * Slave tranmitter. Write '0E' in the 'addr' box of a I2C monitor
;   and the number of bytes to transfer form AVR to I2C monitor.
;   Then press 'Master RX' button. The example transmits
;   always the 0x77 byte, to be seen in the I2C monitor log window.

; * Master transmitter. Push the button "0" in the Control Panel,
;   it will send the countdown sequence 0xFF to 0x01 to the I2C
;   monitor cell "TWO" (slave address = 0x1E).

; * Master receiver. Push the button "1" in the Control Panel,
;   it will address the I2C instance "ONE" (slave address = 0x1D),
;   to get 7 bytes (write such 7 bytes in its "Slave TX" box.
;
; ******************************************************

.include "..\include\m8def.inc"

; ***************************************************
; Interrupt vectors, and Reset
;
reset:
   rjmp start     ; Addr $00
   rjmp INT0_ISR  ; Addr $01   Interrupt from button 0
   rjmp INT1_ISR  ; Addr $02   Interrupt from button 1
   reti           ; Addr $03
   reti           ; Addr $04
   reti           ; Addr $05
   reti           ; Addr $06
   reti           ; Addr $07
   reti           ; Addr $08
   reti           ; Addr $09
   reti           ; Addr $0A
   reti           ; Addr $0B
   reti           ; Addr $0C
   reti           ; Addr $0D
   reti           ; Addr $0E
   reti           ; Addr $0F
   reti           ; Addr $10
   rjmp TWI_ISR   ; Addr $11
   reti           ; Addr $12

; ***************************************************
; INT0 and INT1 services.
;
; INT0 calls the TWI Master TX mode (activated by Key0)
; INT1 calls the TWI Master RX mode (activated by Key1)
;
INT0_ISR:
   rcall MASTER_TX
   reti
INT1_ISR:
   rcall MASTER_RX
   reti


; ***************************************************
; TWI interrupt service routine.
;
; Used to retrieve data in Slave modes. Must check the
; TWSR status, get/send data and clear the interrupt flag
; at the end
;
TWI_ISR:

; Check TWI Status register
; -------------------------
   in r16, TWSR
   andi r16, 0xF8
   cpi r16, 0x80  ; Sataus 0x80 = data received, ACK sent
   breq slave_rx_data
	cpi r16, 0xA8  ; Status 0xA8 = SLA+R received. Will be Slave RX
   breq slave_tx_data
	cpi r16, 0xB8  ; Status 0xB8 = Data has been transmitted ACK OK
	breq slave_tx_data
	cpi r16, 0xC0  ; Status 0xC0 = Data has been transmitted. not ACK,
	rjmp continue  ; Last byte; exit

; Slave receiver
; --------------
slave_rx_data:
   in r16, TWDR   ; Read received byte
   nop            ; .... RX byte is in r16, process it...
   nop            ; place a breakpoint to see the data in
   rjmp continue

; Slave trasmitter
; ---- ------------
slave_tx_data:
	ldi r16, 0x77   ; Load necessary data in TWDR, example: 0x77
	out TWDR, r16		
	; rjmp continue
	
continue:	
   ldi r16, (1<<TWINT)|(1<<TWEA)|(1<<TWEN)|(1<<TWIE) ; Clear flag
   out TWCR, r16
   reti

; **************************************************
; Program starts here after Reset
;
start:
   ldi r16, 0xDF   ; Init stack pointer
   out 0x3D, r16

   ldi r16, 0x0A   ; For INT0, INT1
   out 0x35, r16   ; select falling edge

   ldi r16, 0xC0   ;
   out GIMSK, r16  ; Enable INT0, INT1

   rcall SLAVE_RX  ; Init TWI as SLAVE RX
   sei
forever:
   rjmp forever    ; Infinite loop; all is driven
                   ; by interrupts:
                   ;
                   ; INT0: (Button 0) initiates a TWI master TX
                   ; INT1: (Button 1) initiates a TWI master RX
                   ; Use the I2C monitor cell 'Master TX' button to
                   ; initiate a IC2_monitor --> AVR TWI transmission
                   ; using TWI as Slave Receiver


; **************************************************
; Slave receiver initialization
;
; Initialize by default as SLAVE RX. with slave
; address 0x0E. Click on any I2C monitor cell
; 'Master TX' button with the address set as 0x0E.
;
SLAVE_RX:

   sei             ;         ; This mode is interrupt based
   ldi r16, (0x0E << 1) | 0  ; Own slave address, no GCE
   out TWAR, r16
   ldi r16, (1<<TWEA)|(1<<TWEN)|(1<<TWIE)
   out TWCR, r16
   ret

; **************************************************
; Master receiver
;
; This will read 7 bytes of the I2C monitor
; cell with slave address 0x1D. Such cell will be
; addressed, and it will put into the bus the content
; of the "Slave TX" field
;
MASTER_RX:

   cli   ; Interrupts not used; must poll TWINT bit

; Set bitrate for master mode
; ---------------------------
   ldi r16, 11    ; Minimum value must be 10 !
   out TWBR, r16  ; See formula in Atmel manual

; Send start and wait for TWINT flag
; -----------------------------------
   ldi r16, (1<<TWINT)|(1<<TWSTA)|(1<<TWEN)
   out TWCR, r16
   rcall wait_twint

; Load SLA + R  and clear int flag
; --------------------------------
   ldi r16, (0x1D << 1) | 1 ; Slave address = 0x1D; 1 = Read (RX)
   out TWDR, r16
   ldi r16, (1<<TWINT)|(1<<TWEN)
   out TWCR, r16

; Read loop for 7 bytes
; ---------------------
   ldi r18, 7
read_loop:             ;
   rcall wait_twint    ; Wait for completion
   in r16, TWDR        ; Read received byte

   nop ; ......Received data in r16, process it......
   nop ; place a breakpoint here to see the data

   ldi r16, (1<<TWINT)|(1<<TWEN)|(1<<TWEA) ; Enable aknowledge
   out TWCR, r16
   dec r18
   brne read_loop

   ldi r16, (1<<TWEN)  ; Disable aknowledge for last byte
   out TWCR, r16
   rcall wait_twint    ;

; Send STOP and return
; --------------------
   ldi r16, (1<<TWINT)|(1<<TWEN)|(1<<TWSTO)
   out TWCR, r16
   ret

; **************************************************
; Master transmitter
;
; This will send 15 bytes, codes from 0x0F to 0x01, to
; the I2C monitor cell with slave address 0x1E.
;
MASTER_TX:

   cli   ; Interrupts not used; must poll TWINT bit

; Set bitrate for master mode
; ---------------------------
   ldi r16, 13    ; Minimum value must be 10 !
   out TWBR, r16  ; See formula in Atmel manual

; Send start and wait for TWINT flag
; ----------------------------------
   ldi r16, (1<<TWINT) | (1<<TWSTA) | (1<<TWEN)
   out TWCR, r16
   rcall wait_twint

; Load SLA + W and clear int flag
; -------------------------------
   ldi r16, (0x1E << 1) | 0 ; Address = 0x1E, Write (TX)
   out TWDR, r16
   ldi r16, (1<<TWINT) | (1<<TWEN)
   out TWCR, r16
   rcall wait_twint

; Check TWI Status register
; -------------------------
   in r16, TWSR
   andi r16, 0xF8
   cpi r16, 0x18 ; Status code must be 0x18
   brne ERROR

   ldi r18, 0x0F  ; Init byte to transmit
write_loop:

; Load data into TWDR register
; ----------------------------
   mov r16, r18    ; r18 holds the code to transfer
   out TWDR, r16
   ldi r16, (1<<TWINT)|(1<<TWEN)
   out TWCR, r16
   rcall wait_twint    ; Wait for completion

; Check TWI Status register
; -------------------------
   in r16, TWSR
   andi r16, 0xF8
   cpi r16, 0x28 ; Value must be 0x28
   brne ERROR

  dec r18          ; Decrement data for
  brne write_loop  ; next cycle

; Send STOP and return
; --------------------
   ldi r16, (1<<TWINT) | (1<<TWEN) | (1<<TWSTO)
   out TWCR, r16
   ret

; *********************************************
; Error processing. TWI provides a status
; code, upon the completion of any operation
;
ERROR:
   nop
   nop       ; Some error: not acknowledge, etc.
   nop       ; Process it here.
   nop       ;
   rjmp ERROR


; **********************************************
; Wait for TWI operation, by polling TWINT flag
;
wait_twint:
   in r16, TWCR
   sbrs r16, TWINT
   rjmp wait_twint
   ret











